#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <bergenServer.h>
#include <bergenSample.h>
#include "Bees.h"
#include "cray.h"

#define SPEED 2.0f

/* modes */
#define HANG_OUT	1
#define GO_FLOWER	2
#define POLLENATE	3
#define GO_HIVE		4
#define ANGRY		5

#define SWARM_DISTANCE_SQ	1600.0f

/* values for hithive_ */
#define NOT_HIT		1
#define HIT		2
#define HIT_PLAYED	3

#define MAX_AUDIBLE_DISTSQ	400.0
#define MAX_AMPLITUDE		0.15
#define MAX_AMPLITUDE_PER_BEE	0.025

static float DistanceSq(float *a,float *b);

Bees::Bees(char *options,dcuTextureSet *texset,int usetex)
{
 char *token,*fname;
 int i;
 center_[0] = center_[1] = center_[2] = 0;
 *radius_ = 1;
 numbees_ = 0;
 xsize_ = ysize_ = .2;
 numflowers_ = 0;
 soundfile_ = NULL;
 hitsoundfile_ = NULL;
 color_ = 0xffffffff;
 cullable_ = 0;
 note_ = NULL;
 hitnote_ = NULL;
 hithive_ = NOT_HIT;
 fname = dcuNextToken(&options);
 if (usetex)
	{
	if (fname)
		texture_ = texset->getTexture(fname);
	else
		fprintf(stderr,"CRAY Bees constructor ERROR: missing filename\n");
	}
 while (token = dcuNextToken(&options))
	ParseOption(token);
 mode_ = (int *) CAVEMalloc(numbees_*sizeof(int));
 time_ = (float *) CAVEMalloc(numbees_*sizeof(float));
 x_ = (float *) CAVEMalloc(numbees_*sizeof(float));
 y_ = (float *) CAVEMalloc(numbees_*sizeof(float));
 z_ = (float *) CAVEMalloc(numbees_*sizeof(float));
 ry_ = (float *) CAVEMalloc(numbees_*sizeof(float));
 phase_ = (float *) CAVEMalloc(numbees_*sizeof(float));
 dest_ = (int *) CAVEMalloc(numbees_*sizeof(int));
 delta_ = (struct _Bees_delta *) CAVEMalloc(numbees_*sizeof(struct _Bees_delta));
 for (i=0; i<numbees_; i++)
	{
	mode_[i] = HANG_OUT;
	time_[i] = random() % 6;
	x_[i] = hive_[0] + drand48()+drand48()-1;
	y_[i] = hive_[1] + drand48()+drand48()-1;
	z_[i] = hive_[2] + drand48()+drand48()-1;
	ry_[i] = random() % 360;
	}
}

void Bees::ParseOption(char *optionstr)
{
 enum _option { OPT_HIVEPOS, OPT_COLOR, OPT_FLOWERPOS, OPT_NUM, OPT_SOUND,
		OPT_HITSOUND, OPT_SIZE } opttype;
 struct _dcukeyword opt_table[] = {
	{ "num", OPT_NUM },
	{ "number", OPT_NUM },
	{ "hive", OPT_HIVEPOS },
	{ "hivepos", OPT_HIVEPOS },
	{ "hiveposition", OPT_HIVEPOS },
	{ "flower", OPT_FLOWERPOS },
	{ "flowerpos", OPT_FLOWERPOS },
	{ "flowerposition", OPT_FLOWERPOS },
	{ "color", OPT_COLOR },
	{ "sound", OPT_SOUND },
	{ "soundfile", OPT_SOUND },
	{ "hitsound", OPT_HITSOUND },
	{ "hitsoundfile", OPT_HITSOUND },
	{ "size", OPT_SIZE },
	{ NULL, -1 }
	};
 char *eq,*value;
 eq = strchr(optionstr,'=');
 if (!eq) return;
 *eq = '\0';
 value = eq+1;
 opttype = (enum _option) dcuGetStringID(optionstr,opt_table);
 switch (opttype)
	{
	case OPT_NUM:
		sscanf(value,"%d",&numbees_);
		break;
	case OPT_HIVEPOS:
		sscanf(value,"%f,%f,%f",&hive_[0],&hive_[1],&hive_[2]);
		break;
	case OPT_FLOWERPOS:
		if (numflowers_ >= BEES_MAX_FLOWERS) break;
		sscanf(value,"%f,%f,%f",&flower_[numflowers_][0],
			&flower_[numflowers_][1],&flower_[numflowers_][2]);
		numflowers_++;
		break;
	case OPT_SIZE:
		sscanf(value,"%f,%f",&xsize_,&ysize_);
		break;
	case OPT_COLOR:
		sscanf(value,"%x",&color_);
		break;
	case OPT_SOUND:
		soundfile_ = strdup(value);
		break;
	case OPT_HITSOUND:
		hitsoundfile_ = strdup(value);
		break;
	default:
		fprintf(stderr,"CRAY Bees ERROR: Unknown option \"%s\"\n",
			optionstr);
	}
}

void Bees::GLInitialize(void)
{
 float v[4][3]={{0,-.1,-.1},{0,.1,-.1},{0,-.1,.1},{0,.1,.1}};
 float t[4][2]={{0,0},{0,1},{1,0},{1,1}};
 v[0][1] = -ysize_/2.0f;
 v[0][2] = -xsize_/2.0f;
 v[1][1] = ysize_/2.0f;
 v[1][2] = -xsize_/2.0f;
 v[2][1] = -ysize_/2.0f;
 v[2][2] = xsize_/2.0f;
 v[3][1] = ysize_/2.0f;
 v[3][2] = xsize_/2.0f;
 gl_obj_ = glGenLists(1);
 glNewList(gl_obj_,GL_COMPILE);
   if (!texture_) glColor4ubv((GLubyte*)&color_);
   glBegin(GL_TRIANGLE_STRIP);
     if (texture_) glTexCoord2fv(t[0]);
     glVertex3fv(v[0]);
     if (texture_) glTexCoord2fv(t[1]);
     glVertex3fv(v[1]);
     if (texture_) glTexCoord2fv(t[2]);
     glVertex3fv(v[2]);
     if (texture_) glTexCoord2fv(t[3]);
     glVertex3fv(v[3]);
   glEnd();
 glEndList();
}

void Bees::Render(void)
{
 int i;
 for (i=0; i<numbees_; i++)
    if (mode_[i] != HANG_OUT)
	{
	glPushMatrix();
	  glTranslatef(x_[i],y_[i],z_[i]);
	  glRotatef(ry_[i],0.0,1.0,0.0);
	  glCallList(gl_obj_);
	glPopMatrix();
	}
}

void Bees::Update(float dt)
{
 int b;
 CheckUser();
 for (b=0; b<numbees_; b++)
	switch (mode_[b])
		{
		case HANG_OUT: HangOut(b,dt); break;
		case GO_FLOWER: GoFlower(b,dt); break;
		case POLLENATE: Pollenate(b,dt); break;
		case GO_HIVE: GoHive(b,dt); break;
		case ANGRY: Swarm(b,dt); break;
		}
}

void Bees::CheckUser(void)
{
 float wandpos[3],distsq;
 CAVEGetPosition(CAVE_WAND_NAV,wandpos); 
 distsq = DistanceSq(wandpos,hive_);
 if (distsq < 2)
	{
	int b;
	for (b=0; b<numbees_; b++)
		{
		mode_[b] = ANGRY;
		delta_[b].dx = drand48()+drand48()-1.0f;
		delta_[b].dy = drand48()+drand48()-1.0f;
		delta_[b].dz = drand48()+drand48()-1.0f;
		}
	if (hithive_ == NOT_HIT)
		hithive_ = HIT;
	}
 else
	hithive_ = NOT_HIT;
}

void Bees::HangOut(int b,float dt)
{
 time_[b] -= dt;
 if (time_[b] <= 0)
	{
	mode_[b] = GO_FLOWER;
	dest_[b] = random() % numflowers_;
	}
 else
	{
	float dx,dz;
	x_[b] += (drand48()+drand48()-1.0f) * SPEED * dt;
	y_[b] += (drand48()+drand48()-1.0f) * SPEED * dt;
	z_[b] += (drand48()+drand48()-1.0f) * SPEED * dt;
	dx = hive_[0] - x_[b];
	dz = hive_[2] - z_[b];
	ry_[b] = -atan2f(dx,-dz) * 180.0/M_PI;
	}
}

void Bees::GoFlower(int b,float dt)
{
 float dx,dy,dz,d;
 dx = flower_[dest_[b]][0] - x_[b];
 dy = flower_[dest_[b]][1] - y_[b];
 dz = flower_[dest_[b]][2] - z_[b];
 d = dx*dx + dy*dy + dz*dz;
 if (d < .25)
	{
	mode_[b] = POLLENATE;
	time_[b] = 3 + drand48() * 8;
	}
 else
	{
	d = 1.0f/sqrtf(d);
	dx *= d;
	dy *= d;
	dz *= d;
	x_[b] += (dx+(drand48()-0.5)*3) * SPEED * dt;
	y_[b] += (dy+(drand48()-0.5)*3) * SPEED * dt;
	z_[b] += (dz+(drand48()-0.5)*3) * SPEED * dt;
	ry_[b] = -atan2f(dx,-dz) * 180.0/M_PI;
	}
}

void Bees::Pollenate(int b,float dt)
{
 time_[b] -= dt;
 if (time_[b] <= 0)
	mode_[b] = GO_HIVE;
 else
	{
	float dx,dz;
	x_[b] += (drand48()+drand48()-1.0f) * SPEED/2.0f * dt;
	y_[b] += (drand48()+drand48()-1.0f) * SPEED/2.0f * dt;
	z_[b] += (drand48()+drand48()-1.0f) * SPEED/2.0f * dt;
	dx = flower_[dest_[b]][0] - x_[b];
	dz = flower_[dest_[b]][2] - z_[b];
	ry_[b] = -atan2f(dx,-dz) * 180.0/M_PI;
	}
}

void Bees::GoHive(int b,float dt)
{
 float dx,dy,dz,d;
 dx = hive_[0] - x_[b];
 dy = hive_[1] - y_[b];
 dz = hive_[2] - z_[b];
 d = dx*dx + dy*dy + dz*dz;
 if (d < 1)
	{
	mode_[b] = HANG_OUT;
	time_[b] = drand48() * 8;
	}
 else
	{
	d = 1.0f/sqrtf(d);
	dx *= d;
	dy *= d;
	dz *= d;
	x_[b] += (dx+drand48()-0.5) * SPEED * dt;
	y_[b] += (dy+drand48()-0.5) * SPEED * dt;
	z_[b] += (dz+drand48()-0.5) * SPEED * dt;
	ry_[b] = -atan2f(dx,-dz) * 180.0/M_PI;
	}
}

void Bees::Swarm(int b,float dt)
{
 float dx,dy,dz,d,userpos[3];
 CAVEGetPosition(CAVE_HEAD_NAV,userpos);
 dx = hive_[0] - userpos[0];
 dy = hive_[1] - userpos[1];
 dz = hive_[2] - userpos[2];
 d = dx*dx + dy*dy + dz*dz;
 if (d > SWARM_DISTANCE_SQ)
	{
	for (b=0; b<numbees_; b++)
		mode_[b] = GO_HIVE;
	return;
	}
 dx = userpos[0]+delta_[b].dx - x_[b];
 dy = userpos[1]+delta_[b].dy - y_[b];
 dz = userpos[2]+delta_[b].dz - z_[b];
 if (d < .2)
	{
	x_[b] += (drand48()+drand48()-1.0f) * SPEED * dt;
	y_[b] += (drand48()+drand48()-1.0f) * SPEED * dt;
	z_[b] += (drand48()+drand48()-1.0f) * SPEED * dt;
	}
 else
	{
	d = 1.0f/sqrtf(d);
	dx *= d;
	dy *= d;
	dz *= d;
	x_[b] += (dx+drand48()-0.5) * SPEED*3.0f * dt;
	y_[b] += (dy+drand48()-0.5) * SPEED*3.0f * dt;
	z_[b] += (dz+drand48()-0.5) * SPEED*3.0f * dt;
	}
 ry_[b] = -atan2f(dx,-dz) * 180.0/M_PI + (drand48()+drand48()-1.0f)*15.0f;
}

void Bees::InitSound(bergenServer *server)
{
 if ((soundfile_) && (server))
	{
	note_ = new bergenSample(soundfile_,server);
	note_->setLoop(1);
	note_->setAmplitude(0.0);
	note_->play();
	ampl_ = 0;
	pitch_ = 1;
	}
 if ((hitsoundfile_) && (server))
	{
	hitnote_ = new bergenSample(hitsoundfile_,server);
	hitnote_->setAmplitude(0.75);
	}
}

void Bees::UpdateSound(float *userpos)
{
 float dx,dy,dz,distsq,newampl=0,multiplier=1.0,newpitch=1;
 int b;
 if (hithive_ == HIT)
	{
	if (hitnote_)
		hitnote_->play();
	hithive_ = HIT_PLAYED;
	}
 if (!note_)
	return;
 if (mode_[0] == ANGRY)
	{
	multiplier = 3.0;
	newpitch = 2.0;
	}
 for (b=0; b < numbees_; b++)
	{
	dx = x_[b]-userpos[0];
	dy = y_[b]-userpos[1];
	dz = z_[b]-userpos[2];
	distsq = dx*dx+dy*dy+dz*dz;
	if (distsq < MAX_AUDIBLE_DISTSQ-1.0)
		newampl += MAX_AMPLITUDE_PER_BEE * multiplier *
			(MAX_AUDIBLE_DISTSQ-distsq) / MAX_AUDIBLE_DISTSQ;
	}
 if (newampl > MAX_AMPLITUDE*multiplier)
	newampl = MAX_AMPLITUDE*multiplier;
 if (fabs(ampl_ - newampl) > .0025)
	{
	note_->setAmplitude(newampl);
	ampl_ = newampl;
	}
/*  This can be included whenever bergen actually supports pitch changes */
/***
 if (newpitch != pitch_)
	{
	note_->setPitch(newPitch);
	pitch_ = newpitch;
	}
***/
}

void Bees::EndSound(void)
{
 if (note_)
	{
	note_->stop();
	delete note_;
	note_ = NULL;
	}
 if (hitnote_)
	{
	hitnote_->stop();
	delete hitnote_;
	hitnote_ = NULL;
	}
}


static float DistanceSq(float *a,float *b)
{
 float dx,dy,dz;
 dx = a[0] - b[0];
 dy = a[1] - b[1];
 dz = a[2] - b[2];
 return dx*dx + dy*dy + dz*dz;
}
