#include <string.h>
#ifdef __sgi
#include <bstring.h>
#endif
#include "cray.h"
#include "PictObj.h"
#include "util.h"
#include <dave.h>

#define FRICTION_DAMPING  .1f
#define GRAVITY		32.0f
#define MAX_VEL_SQ	100.0f

#define TRANSFORM(i,j)  transform_[i*4+j]

PictureObject::PictureObject(char *options,dcuTextureSet *texset,int usetex) :
    Picture(options,texset,usetex)
{
 grabbed_ = (int *) CAVEMalloc(sizeof(int));
 *grabbed_ = 0;
 transform_ = (float *) CAVEMalloc(16*sizeof(float));
 bzero((void *)transform_,16*sizeof(float));
 TRANSFORM(0,0) = TRANSFORM(1,1) = TRANSFORM(2,2) = TRANSFORM(3,3) = 1;
 TRANSFORM(3,0) = x_;
 TRANSFORM(3,1) = y_;
 TRANSFORM(3,2) = z_;
 prevt_ = 0;
 vx_ = vy_ = vz_ = 0;
 x_ = y_ = z_ = 0;
 cullable_ = 0;
}

void PictureObject::GLInitialize(void)
{
 Picture::GLInitialize();
}

void PictureObject::Render(void)
{
 float transf[16];
 glPushMatrix();
  if (*grabbed_)
	{
	float wandmat[16];
	GetWandMatrix(wandmat);
	glMultMatrixf(wandmat);
	}
  bcopy((void *)transform_,(void *)transf,16*sizeof(float));
  glMultMatrixf(transf);
  glCallList(gl_obj_);
 glPopMatrix();
}

void PictureObject::Update(float dt)
{
/* Use button1 or second finger of glove (in wand-n-glove config) to grab */
 if (((CAVEBUTTON1) ||
      ((CAVEController->num_valuators > 3) && (CAVEController->valuator[3] > 0.65)))
     && (!*grabbed_) && (WandTouching()))
	{
	float wandmat[16],invwandmat[16];
	Matrix wandMatrix, invwandMatrix;
	GetWandMatrix(wandmat);
	memcpy(wandMatrix,wandmat,sizeof(Matrix));
	dpInvertAffineMatrix(wandMatrix,invwandMatrix);
	memcpy(invwandmat,invwandMatrix,sizeof(Matrix));
	*grabbed_ = 1;
	MatrixMultiply(invwandmat,(float *)transform_,(float *)transform_);
	}
 else if (((!CAVEBUTTON1) &&
	  (!((CAVEController->num_valuators > 3) && (CAVEController->valuator[3] > 0.65))))
	   && (*grabbed_))
	{
	float wandmat[16];
	GetWandMatrix(wandmat);
	*grabbed_ = 0;
	MatrixMultiply(wandmat,(float *)transform_,(float *)transform_);
	}
 if (dt>0)
	{
	if (*grabbed_)
		GetVelocity(dt);
	else
		Move(dt);
	}
}

void PictureObject::GetVelocity(float dt)
{
 float x,y,z,t,wandpos[3],v,clamp;
 t = CAVEGetTime();
 if (t-prevt_ < .25) return;
 CAVEGetPosition(CAVE_WAND_NAV,wandpos); 
 x = wandpos[0];
 y = wandpos[1];
 z = wandpos[2];
 dt = t-prevt_;
 vx_ = (x - prevx_)/dt * 5;
 vy_ = (y - prevy_)/dt * 5;
 vz_ = (z - prevz_)/dt * 5;
 v = vx_*vx_ + vy_*vy_ + vz_*vz_;
 if (v > MAX_VEL_SQ)
	{
	clamp = sqrtf(MAX_VEL_SQ/v);
	vx_ *= clamp;
	vy_ *= clamp;
	vz_ *= clamp;
	}
 prevx_ = x;
 prevy_ = y;
 prevz_ = z;
 prevt_ = t;
}

void PictureObject::Move(float dt)
{
 float x,y,z;
 prevx_ = TRANSFORM(3,0);
 prevy_ = TRANSFORM(3,1);
 prevz_ = TRANSFORM(3,2);
 x = prevx_ + vx_ * dt;
 y = prevy_ + vy_ * dt;
 z = prevz_ + vz_ * dt;
 if (y < 0)
	{
	y = 0;
	vy_ = 0;
	vx_ *= FRICTION_DAMPING*dt;
	vz_ *= FRICTION_DAMPING*dt;
	}
 TRANSFORM(3,0) = x;
 TRANSFORM(3,1) = y;
 TRANSFORM(3,2) = z;
 if (y > 0)
	vy_ -= GRAVITY * dt;
}

int PictureObject::WandTouching(void)
{
 float wandpos[3],dx,dy,dz;
 CAVEGetPosition(CAVE_WAND_NAV,wandpos); 
 dx = wandpos[0] - TRANSFORM(3,0) - center_[0];  /* This is wrong - must multiply */
 dy = wandpos[1] - TRANSFORM(3,1) - center_[1];  /*  center_ by transform_  */
 dz = wandpos[2] - TRANSFORM(3,2) - center_[2];
 return ((dx*dx+dy*dy+dz*dz) < ((*radius_)*(*radius_)));
}

void PictureObject::GetWandMatrix(float * m)
{
 float xrotmat[] = {1,0,0,0, 0,1,0,0, 0,0,1,0, 0,0,0,1};
 float yrotmat[] = {1,0,0,0, 0,1,0,0, 0,0,1,0, 0,0,0,1};
 float zrotmat[] = {1,0,0,0, 0,1,0,0, 0,0,1,0, 0,0,0,1};
 float or[3],rx,ry,rz,pos[3];
 CAVEGetPosition(CAVE_WAND_NAV,pos); 
 CAVEGetOrientation(CAVE_WAND_NAV,or); 
 rx = DTOR(or[0]);
 ry = -DTOR(or[1]);
 rz = DTOR(or[2]);
 bzero((void *)m,64);
 m[0] = m[5] = m[10] = m[15] = 1;
 m[12] = pos[0];
 m[13] = pos[1];
 m[14] = pos[2];
 yrotmat[0*4+0] = yrotmat[2*4+2] = cosf(ry);
 yrotmat[0*4+2] = sinf(ry);
 yrotmat[2*4+0] = -yrotmat[0*4+2];
 xrotmat[1*4+1] = xrotmat[2*4+2] = cosf(rx);
 xrotmat[1*4+2] = sinf(rx);
 xrotmat[2*4+1] = -xrotmat[1*4+2];
 zrotmat[0*4+0] = zrotmat[1*4+1] = cosf(rz);
 zrotmat[0*4+1] = sinf(rz);
 zrotmat[1*4+0] = -zrotmat[0*4+1];
 MatrixMultiply(m,yrotmat,m);
 MatrixMultiply(m,xrotmat,m);
 MatrixMultiply(m,zrotmat,m);
}

void PictureObject::MatrixMultiply(float * a,float * b,float * out)
{
 int i,j,k;
 Matrix result;
 float sum;
 for (i=0; i<4; i++)
    for (j=0; j<4; j++)
	{
	sum = 0;
	for (k=0; k<4; k++)
		sum += b[i*4+k] * a[k*4+j];
	result[i][j] = sum;
	}
 for (i=0; i<4; i++)
    for (j=0; j<4; j++)
	out[i*4+j] = result[i][j];
}
