#include <stdio.h>
#include <math.h>
#include <string.h>
#include "Picture.h"
#include "util.h"

Picture::Picture(char *options,dcuTextureSet *texset,int usetex)
{
 char *fname,*token;
 InitValues();
 fname = dcuNextToken(&options);
 if (usetex)
	{
	if (fname)
		texture_ = texset->getTexture(fname);
	else
		fprintf(stderr,"CRAY Picture constructor ERROR: missing filename\n");
	}
 while (token = dcuNextToken(&options))
	ParseOption(token);
}

void Picture::InitValues(void)
{
 x_ = y_ = z_ = 0;
 rx_ = ry_ = rz_ = 0;
 xscale_ = yscale_ = 1;
 center_[0] = center_[1] = center_[2] = 0;
 *radius_ = .5;
 color_ = 0xffffffff;
 texc_ll_[0] = texc_ll_[1] = 0;
 texc_ur_[0] = texc_ur_[1] = 1;
 cullable_ = 1;
}

void Picture::ParseOption(char *optionstr)
{
 enum _option { OPT_POSITION, OPT_ORIENTATION, OPT_SIZE, OPT_COLOR, OPT_TEXC } opttype;
 struct _dcukeyword opt_table[] = {
	{ "position", OPT_POSITION },
	{ "pos", OPT_POSITION },
	{ "orientation", OPT_ORIENTATION },
	{ "orient", OPT_ORIENTATION },
	{ "angle", OPT_ORIENTATION },
	{ "size", OPT_SIZE },
	{ "scale", OPT_SIZE },
	{ "color", OPT_COLOR },
	{ "texc", OPT_TEXC },
	{ "texcoords", OPT_TEXC },
	{ NULL, -1 }
	};
 char *eq,*value;
 eq = strchr(optionstr,'=');
 if (!eq) return;
 *eq = '\0';
 value = eq+1;
 opttype = (enum _option) dcuGetStringID(optionstr,opt_table);
 switch (opttype)
	{
	case OPT_POSITION:
		sscanf(value,"%f,%f,%f",&x_,&y_,&z_);
		break;
	case OPT_ORIENTATION:
		sscanf(value,"%f,%f,%f",&rx_,&ry_,&rz_);
		rx_ = DTOR(rx_);
		ry_ = DTOR(ry_);
		rz_ = DTOR(rz_);
		break;
	case OPT_SIZE:
		sscanf(value,"%f,%f",&xscale_,&yscale_);
		break;
	case OPT_COLOR:
		sscanf(value,"%x",&color_);
		break;
	case OPT_TEXC:
		sscanf(value,"%f,%f,%f,%f",&texc_ll_[0],&texc_ll_[1],
			&texc_ur_[0],&texc_ur_[1]);
		break;
	default:
		fprintf(stderr,"CRAY Picture ERROR: Unknown option \"%s\"\n",
			optionstr);
	}
}

void Picture::GLInitialize(void)
{
 float v0[3],v1[3],v2[3],v3[3];
 float t0[2]={0,0},t1[2]={1,0},t2[2]={0,1},t3[2]={1,1};
/* Added 5/16/96 - texcoords, to allow clipping out parts of larger images */
 t0[0] = texc_ll_[0];   t0[1] = texc_ll_[1];
 t1[0] = texc_ur_[0];   t1[1] = texc_ll_[1];
 t2[0] = texc_ll_[0];   t2[1] = texc_ur_[1];
 t3[0] = texc_ur_[0];   t3[1] = texc_ur_[1];
 gl_obj_ = glGenLists(1);
 glNewList(gl_obj_,GL_COMPILE);
   if (!texture_) glColor4ubv((GLubyte*)&color_);
   glBegin(GL_TRIANGLE_STRIP);
     if (texture_) glTexCoord2fv(t0);
     v0[0] = x_ - xscale_/2 * (cos(ry_)*cos(rz_)+sin(ry_)*sin(rx_)*sin(rz_));
     v0[1] = y_ - xscale_/2 * cos(rx_)*sin(rz_);
     v0[2] = z_ - xscale_/2 * (-sin(ry_)*cos(rz_)+cos(ry_)*sin(rx_)*sin(rz_));
     glVertex3fv(v0);
     if (texture_) glTexCoord2fv(t1);
     v1[0] = v0[0] + xscale_ * (cos(ry_)*cos(rz_)+sin(ry_)*sin(rx_)*sin(rz_));
     v1[1] = v0[1] + xscale_ * cos(rx_)*sin(rz_);
     v1[2] = v0[2] + xscale_ * (-sin(ry_)*cos(rz_)+cos(ry_)*sin(rx_)*sin(rz_));
     glVertex3fv(v1);
     if (texture_) glTexCoord2fv(t2);
     v2[0] = v0[0] + yscale_ * (-cos(ry_)*sin(rz_) + sin(ry_)*sin(rx_)*cos(rz_));
     v2[1] = v0[1] + yscale_ * (cos(rz_)*cos(rx_));
     v2[2] = v0[2] + yscale_ * (sin(ry_)*sin(rz_) + cos(ry_)*sin(rx_)*cos(rz_));
     glVertex3fv(v2);
     if (texture_) glTexCoord2fv(t3);
     v3[0] = v2[0] + xscale_ * (cos(ry_)*cos(rz_)+sin(ry_)*sin(rx_)*sin(rz_));
     v3[1] = v2[1] + xscale_ * cos(rx_)*sin(rz_);
     v3[2] = v2[2] + xscale_ * (-sin(ry_)*cos(rz_)+cos(ry_)*sin(rx_)*sin(rz_));
     glVertex3fv(v3);
   glEnd();
 glEndList();
 center_[0] = (v0[0] + v1[0] + v2[0] + v3[0]) / 4;
 center_[1] = (v0[1] + v1[1] + v2[1] + v3[1]) / 4;
 center_[2] = (v0[2] + v1[2] + v2[2] + v3[2]) / 4;
 *radius_ = sqrt(xscale_*xscale_/4.0 + yscale_*yscale_/4.0);
}

void Picture::Render(void)
{
 glCallList(gl_obj_);
}

void Picture::Update(float)
{
}
