#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <math.h>
#include <malloc.h>
#include <sys/types.h>
#include "cray.h"
#include "dcuTextureSet.h"
#include "Bees.h"
#include "Flies.h"
#include "Butterfly.h"
#include "Picture.h"
#include "PictObj.h"
#include "util.h"


extern "C" { int CheckForExitCommand(void); }


extern	void	cray_InitAudio(world_t *);
extern	void	cray_UpdateAudio(world_t *);
extern	void	cray_StopAudio(world_t *);
static	world_t *cray_InitWorld(int argc,char **argv,char *configFile);
static	void	cray_InitGL(world_t *world);
static	void	cray_DrawWorld(world_t *);
static	void	cray_Update(world_t *world);
static	void	load_world(world_t *);
static	void	draw_elements(world_t *world);
static	void	parse_command_line(int argc,char **argv,world_t *world);
static	void	read_config_file(char *configFile,world_t *world);
static	void	InitNavigation(world_t *);
static	void	DoNavigation(float,world_t *);
static	int	cull_element(SceneElement *p);

static dcuTextureSet texture_set;

main(int argc,char **argv)
{
 world_t *world;
 CAVESetOption(CAVE_GL_SAMPLES,8);
 CAVEConfigure(&argc,argv,NULL);
 world = cray_InitWorld(argc,argv,"Cray.config");
 if (world->do_audio)
	cray_InitAudio(world);
 CAVEInit();
 CAVEInitApplication((CAVECALLBACK)cray_InitGL,1,world);
 CAVEDisplay((CAVECALLBACK)cray_DrawWorld,1,world);
 while ((!CAVEgetbutton(CAVE_ESCKEY)) /*&& (!CheckForExitCommand())*/ )
	{
	cray_Update(world);
	if (world->do_audio)
		cray_UpdateAudio(world);
	CAVEDisplaySync();
	if ((world->frameRateTest) && (CAVEGetTime() > 15))
		{
		printf("%.2f fps\n",*CAVEFramesPerSecond);
		break;
		}
	}
 if (world->do_audio)
	cray_StopAudio(world);
 CAVEExit();
 return 0;
}

static void cray_Update(world_t *world)
{
 static float prevtime=-1;
 float curtime,dt;
 int i;
 if (CAVEDistribMaster())
	{
	if (prevtime < 0)
		prevtime = CAVEGetTime();
	curtime = CAVEGetTime();
	dt = curtime - prevtime;
	prevtime = curtime;
	DoNavigation(dt,world);
	for (i=0; i<world->num_elements; i++)
		world->elem[i]->Update(dt);
	}
}

static world_t * cray_InitWorld(int argc,char **argv,char *configFile)
{
 world_t *world;
 world = (world_t *) CAVEMalloc(sizeof(world_t));
 memset(world,0,sizeof(world_t));
 world->startPos[0] = world->startPos[1] = world->startPos[2] = 0;
 world->startHeading = 0;
 if (argc>1)
	parse_command_line(argc,argv,world);
 else
	read_config_file(configFile,world);
 if (world->world_descr_file)
	load_world(world);
 InitNavigation(world);
 return world;
}

static void read_config_file(char *configFile,world_t *world)
{
 FILE *fp;
 char line[1024],*dir;
 if (!(fp = fopen(configFile,"r")))
	{
	perror(configFile);
	return;
	}
 while (fgets(line,1024,fp))
	{
	enum _opt { OPT_TEX, OPT_AUDIO, OPT_WORLD, OPT_DIR, OPT_HAND,
			OPT_STARTPOS, OPT_STARTHEADING, OPT_TEST } opt;
	struct _dcukeyword opt_table[] = {
			{ "texture", OPT_TEX },
			{ "tex", OPT_TEX },
			{ "audio", OPT_AUDIO },
			{ "world", OPT_WORLD },
			{ "dir", OPT_DIR },
			{ "directory", OPT_DIR },
			{ "hand", OPT_HAND },
			{ "startpos", OPT_STARTPOS },
			{ "startposition", OPT_STARTPOS },
			{ "startheading", OPT_STARTHEADING },
			{ "testframerate", OPT_TEST },
			{ NULL, -1 }
			};
	char *s = line,*optstr;
	optstr = dcuNextToken(&s);
	if ((!optstr) || (*optstr == '#'))
		continue;
	opt = (enum _opt) dcuGetStringID(optstr,opt_table);
	switch (opt)
		{
		case OPT_DIR:
			if (chdir(dir=dcuNextToken(&s)) == -1)
				perror(dir);
			break;
		case OPT_TEX:
			world->use_textures = 1;
			break;
		case OPT_AUDIO:
			world->do_audio = 1;
			world->sound_descr_file = strdup(dcuNextToken(&s));
			break;
		case OPT_WORLD:
			world->world_descr_file = strdup(dcuNextToken(&s));
			break;
		case OPT_HAND:
#ifdef USE_HAND
			world->hand_open = wfReadObject(dcuNextToken(&s));
			world->hand_closed = wfReadObject(dcuNextToken(&s));
#endif
			break;
		case OPT_STARTPOS:
			world->startPos[0] = atof(dcuNextToken(&s));
			world->startPos[1] = atof(dcuNextToken(&s));
			world->startPos[2] = atof(dcuNextToken(&s));
			break;
		case OPT_STARTHEADING:
			world->startHeading = atof(dcuNextToken(&s));
			break;
		case OPT_TEST:
			world->frameRateTest = true;
			break;
		default: fprintf(stderr,"CRAY Warning: Unknown config option \"%s\"\n",
				optstr);
		}
	}
 fclose(fp);
}

static void parse_command_line(int argc,char **argv,world_t *world)
{
 int opt;
 while ((opt = getopt(argc,argv,"a:td:HF")) != EOF)
	{
	if (opt=='a') { world->do_audio=1;
			world->sound_descr_file = strdup(optarg); }
	else if (opt=='t') world->use_textures=1;
	else if (opt=='d') { if (chdir(optarg) == -1) perror(optarg); }
	else if (opt=='H')
		{
#ifdef USE_HAND
		world->hand_open = wfReadObject(argv[optind++]);
		world->hand_closed = wfReadObject(argv[optind++]);
#endif
		}
	else if (opt=='F')
		world->frameRateTest = true;
	}
 if (optind >= argc)
	{
	fprintf(stderr,"Usage: %s [-a sound.descr] [-t] "
		"[-H openhand.obj closedhand.obj] world.descr\n",
			argv[0]);
	exit(1);
	}
 world->world_descr_file = strdup(argv[optind]);
}

static void load_world(world_t *world)
{
 FILE *fp;
 char line[1024],*dir;
 int count;
 char *path=NULL;
 if (!(path=getcwd(NULL,512)))
	perror("CRAY getcwd");
 printf("CRAY: Reading \"%s\"\n",world->world_descr_file);
 if ((fp = fopen(world->world_descr_file,"r")) == NULL)
	{
	perror(world->world_descr_file);
	return;
	}
 world->num_elements = 0;
 count = 0;
 while (fgets(line,1024,fp))
	count++;
 world->elem = (SceneElement **) calloc(count,sizeof(SceneElement *));
 fseek(fp,SEEK_SET,0);
/*
 texture_set.setComponents(GL_RGBA4_EXT);
*/
 texture_set.setMinFilter(GL_LINEAR);
 texture_set.setMagFilter(GL_LINEAR);
#ifdef __sgi
 texture_set.setMipmap(); 
#else
 texture_set.setByteOrder(DCU_RGBA);
#endif
 while (fgets(line,1024,fp))
	{
	enum _type { TYPE_PICTURE, TYPE_BUTTERFLY, TYPE_FLIES, TYPE_PICTOBJ,
			TYPE_BEES, TYPE_DIR } type;
	struct _dcukeyword type_table[] = {
			{ "dir", TYPE_DIR },
			{ "directory", TYPE_DIR },
			{ "picture", TYPE_PICTURE },
			{ "pict", TYPE_PICTURE },
			{ "butterfly", TYPE_BUTTERFLY },
			{ "bees", TYPE_BEES },
			{ "bee", TYPE_BEES },
			{ "flies", TYPE_FLIES },
			{ "fly", TYPE_FLIES },
			{ "pictureobject", TYPE_PICTOBJ },
			{ "pictobj", TYPE_PICTOBJ },
			{ NULL, -1 }
			};
	char *s = line,*typestr;
	typestr = dcuNextToken(&s);
	type = (enum _type) dcuGetStringID(typestr,type_table);
	switch (type) {
		case TYPE_DIR:
			if (chdir(dir=dcuNextToken(&s)) == -1)
				perror(dir);
			break;
		case TYPE_PICTURE:
			world->elem[world->num_elements] = new Picture(s,
					&texture_set,world->use_textures);
			world->num_elements++;
			break;
		case TYPE_PICTOBJ:
			world->elem[world->num_elements] = new PictureObject(s,
					&texture_set,world->use_textures);
			world->num_elements++;
			break;
		case TYPE_BUTTERFLY:
			world->elem[world->num_elements] =
					new Butterfly(s,&texture_set,world->use_textures);
			world->num_elements++;
			break;
		case TYPE_BEES:
			world->elem[world->num_elements] =
					new Bees(s,&texture_set,world->use_textures);
			world->num_elements++;
			break;
		case TYPE_FLIES:
			world->elem[world->num_elements] = new Flies(s);
			world->num_elements++;
			break;
		default:
			fprintf(stderr,"CRAY load_world ERROR: unknown type"
				" \"%s\"\n",typestr);
		}
	}
 fclose(fp);
 if (path)
	{
	chdir(path);
	free(path);
	}
 printf("CRAY: Finished reading \"%s\"\n",world->world_descr_file);
}


static void InitNavigation(world_t *world)
{
 CAVENavLoadIdentity();
 CAVENavTranslate(world->startPos[0],world->startPos[1],world->startPos[2]);
 CAVENavRot(world->startHeading,'y');
}

static void DoNavigation(float dt,world_t *world)
{
 float jx=CAVE_JOYSTICK_X,jy=CAVE_JOYSTICK_Y;
 if ((jx == -1) && (jy == -1))
	return;
 if ((CAVEBUTTON1) && (CAVEBUTTON2) && (CAVEBUTTON3))
	InitNavigation(world);
 if (fabs(jy)>0.2)
	{
	float w[3];
	CAVEGetVector(CAVE_WAND_FRONT,w);
	CAVENavTranslate(w[0]*jy*40.0f*dt,0.0,w[2]*jy*40.0f*dt);
	}
 if (fabs(jx)>0.2)
	CAVENavRot(-jx*90.0f*dt,'y');
 if (CAVEBUTTON3)
	{
	float ori[3],speed,turnspeed;
	CAVEGetOrientation(CAVE_WAND,ori);
	speed = (90.0f - fabs(ori[1])) / 90.0f * 40.0 * dt;
	CAVENavTranslate(0.0, 0.0, -speed);
	if (ori[1] > 90.0f)
		turnspeed = (ori[1] - 180.0f) / 90.0f;
	else if (ori[1] < -90.0f)
		turnspeed = (ori[1] + 180.0f) / 90.0f;
	else
		turnspeed = -ori[1] / 90.0f;
	CAVENavRot(-turnspeed*90.0f*dt,'y');
	}
 if (CAVEgetbutton(CAVE_NKEY))
	{
	float pos[3], ori[3];
	CAVEGetPosition(CAVE_HEAD_NAV,pos);
	CAVEGetOrientation(CAVE_HEAD_NAV,ori);
	printf("Position: %f,%f    Heading: %f\n",pos[0],pos[2],ori[1]);
	}
}

static void cray_InitGL(world_t *world)
{
 int i;
 world->fog = 0;
 CAVEFar = 300.0; CAVENear = 0.1;
 if (world->fog)
	{
	float fogcolor[]={.2196,.4706,1};
	printf("enabling fog\n");
	glEnable(GL_FOG);
	glHint(GL_FOG_HINT,GL_FASTEST);
	glFogi(GL_FOG_MODE,GL_LINEAR);
	glFogfv(GL_FOG_COLOR,fogcolor);
	glFogf(GL_FOG_START,40.0);
	glFogf(GL_FOG_END,90.0);
	}
 glShadeModel(GL_FLAT);
#ifdef __sgi
 world->msalpha = glIsEnabled(GL_MULTISAMPLE_SGIS);
#else
 world->msalpha = 0;
#endif
 if (world->use_textures)
{ printf("defining textures\n");
	texture_set.define();
printf("done\n"); }
 for (i=0; i<world->num_elements; i++)
	world->elem[i]->GLInitialize();
}

static void cray_DrawWorld(world_t *world)
{
 glClearColor(0.22,0.47,1.0,0.0);
 glClear(GL_COLOR_BUFFER_BIT);
 glClear(GL_DEPTH_BUFFER_BIT);
 if (world->hand_open)
	{
	glPushMatrix();
	CAVEWandTransform();
#ifdef USE_HAND
	if (CAVEBUTTON1)
		{
		if (world->hand_closed)
			wfDrawObject(world->hand_closed);
		}
	else
		wfDrawObject(world->hand_open);
#endif
	glPopMatrix();
	}
 CAVENavTransform();
 draw_elements(world);
#if 0
 if (getbutton(DKEY))
	{
	static int num=0;
	char basename[64];
	sprintf(basename,"cray%02d",num);
	num++;
	if (CAVEInStereo())
		dpDumpSgiStereo(basename);
	else
		dpDumpSgi(basename);
	}
#endif
}

static void draw_elements(world_t *world)
{
 int i;
 dcuTexture *boundtex=NULL;
 glColor4f(1.0,1.0,1.0,1.0);
 if (world->use_textures)
	{
	if (world->msalpha)
		{
#ifdef __sgi
		glEnable(GL_SAMPLE_ALPHA_TO_MASK_SGIS);
		glEnable(GL_SAMPLE_ALPHA_TO_ONE_SGIS);
#endif
		}
	else
		{
		glEnable(GL_ALPHA_TEST);
		glAlphaFunc(GL_GEQUAL,0.25);
		}
	for (i=0; i<world->num_elements; i++)
/*	    if (!cull_element(world->elem[i])) */
		{
		if (world->elem[i]->Texture() != boundtex)
			{
			boundtex = world->elem[i]->Texture();
			if (boundtex)
				boundtex->bind();
			else
				glDisable(GL_TEXTURE_2D);
			glColor4f(1.0,1.0,1.0,1.0);
			}
		world->elem[i]->Render();
		}
#ifdef __sgi
	glDisable(GL_SAMPLE_ALPHA_TO_MASK_SGIS);
	glDisable(GL_SAMPLE_ALPHA_TO_ONE_SGIS);
#endif
	glDisable(GL_ALPHA_TEST);
	glAlphaFunc(GL_ALWAYS,0.0);
	glDisable(GL_TEXTURE_2D);
	}
 else
	for (i=0; i<world->num_elements; i++)
	    {
/*	    if (!cull_element(world->elem[i])) */
		world->elem[i]->Render();
	    }
}

static int cull_element(SceneElement *p)
{
 if (!p->Cullable()) return 0;
 float cavepos[3],eye[3],radius = p->Radius();
 if (radius<0)
	return 0;
 CAVEGetPosition(CAVE_HEAD,eye);
 CAVENavConvertWorldToCAVE(p->Center(),cavepos);
 switch (CAVEWall)
	{
	case CAVE_FRONT_WALL:	if (cavepos[2] > eye[2]+radius) return 1;
				break;
	case CAVE_LEFT_WALL:	if (cavepos[0] > eye[0]+radius) return 1;
				break;
	case CAVE_RIGHT_WALL:	if (cavepos[0] < eye[0]+radius) return 1;
				break;
/* Note: FLOOR culling assumes nothing has a y < 0 */
	case CAVE_FLOOR_WALL:	if ((fabs(cavepos[0]) > eye[0]+radius) ||
				    (fabs(cavepos[2]) > eye[2]+radius) ||
				    (cavepos[1] > eye[1]+radius))
					return 1;
				break;
	}
 return 0;
}
